// eased_shapes primary cube functions
// filename: es_cube.scad

/* Copyright 2018-2020 Andrew H. Wakefield

    es_cube.scad is part of the eased_shapes.scad library

    This library is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

module eased_cube (

        // set dimensions, with several options. The
        // actual top and bottom sizes are determined
        // by the highest-priority valid setting (=>0)
        // from size, size_top, and size_bottom; valid
        // dimensions from size_top and size_bottom
        // override corresponding dimensions in size.
        // For z dimension, height has the highest
        // priority, followed by h, then size_top[2],
        // size_bottom[2], and size[2]. size, size_top,
        // and size_bottom are vectors containing
        // [x, y, z]; height and h are scalar.

        size = [-1,-1,-1],
        size_bottom = [-1,-1,-1],
        size_top = [-1,-1,-1],
        h = -1,                   // alternate z dimension
        height = -1,              // another alternate z

        // size of top, bottom, and corner features

        bottom = 0,
        top = 0,
        corners = 0,
                                  
        // bevel, fillet, concave consist of a vector of
        // true/false values for [bottom,top,corners],
        // indicating whether this type of feature is
        // to be used. Note that concave overrides
        // fillet overrides bevel.

        bevel = es_none,
        fillet = es_none,         // outward fillet
        concave = es_none,      // inward fillet

        // top_type, bottom_type, corner_type consist of
        // a vector of true/false values for [bevel, fillet,
        // concave], indicating which type of feature
        // is applied to the top, bottom, or corners.
        // Again, concave overrides fillet overrides bevel
 
        bottom_type = es_none,
        top_type = es_none,
        corner_type = es_none,

        // slope_to_origin applies only when top and
        // bottom sizes differ. If slope_to_origin is true,
        // any slope will occur on the right and back
        // sides; if false any slope will occur equally
        // on left/right and on front/back
        
        slope_to_origin = false,
 
        // if any parameter below is true, the
        // cube is centered on the specified axis/axes.
        // Note that the centering is based on the largest
        // face, top or bottom, including where it winds
        // up as a result of slope_to_origin setting
        
        center = false,
        center_x = false,
        center_y = false,
        center_z = false,
        center_xy = false,
        center_xyz = [false, false, false],

        // if limit_size is set true, size of top /
        // bottom / corner features is limited by the
        // overall dimensions of the cube
        
        limit_size = false,

        // debugging settings; for normal operation,
        // leave these set to true. If enable_hull is
        // false, only the corner objects are shown,
        // without the hull around them. If enable_
        // difference is false, the cylinders used
        // to create concave shapes will be shown
        // rather than cut from the cube

        enable_hull = true,
        enable_difference = true
    )
{
    // height has highest priority for z, followed by
    // h, size_top, size_bottom, then size;
    // if none of these is valid, default=10

    z_size = es_get_priority_parameter (
        [height,h,size_top[2],size_bottom[2],size[2],10] );

    // size_top and size_bottom dimensions for x, y
    // have highest priority, followed by size;
    // if none of these is valid, default=10.
    // z is set to 0 for b_size and to z_size for t_size

    b_size = es_assemble_priority_vector (
        [ [ size_bottom[0], size_bottom[1], 0 ],
          [ size[0], size[1], 0 ], [10,10,0] ] );
    t_size = es_assemble_priority_vector (
        [ [-1,-1,z_size], size_top, size, [10,10,10] ] );

    dx = b_size[0] - t_size[0];
    dy = b_size[1] - t_size[1];

    // calculate the angles for each side

    bx_angle = slope_to_origin ?
        es_get_angle(z_size,dx) :
        es_get_angle(z_size,dx/2);

    by_angle = slope_to_origin ?
        es_get_angle(z_size,dy) :
        es_get_angle(z_size,dy/2);

    is_angled = ( bx_angle != 90 || by_angle != 90 );
    
    angles = // bottom: [ left, right, front, back ]
             // top : [ left, right, front, back ]
        slope_to_origin ?
            [ [ 90, bx_angle,
                90, by_angle ],
              [ 90, 180-bx_angle,
                90, 180-by_angle ] ]
          : [ [ bx_angle, bx_angle,
               by_angle, by_angle ],
              [ 180-bx_angle, 180-bx_angle,
                180-by_angle, 180-by_angle ] ];

    max_b_angle = max(bx_angle,by_angle);
    max_t_angle = max(180-bx_angle,180-by_angle);
    
    no_bevel_b = (max_b_angle > 135);
    no_bevel_t = (max_t_angle > 135);

    b_type = es_set_type(bevel[0],fillet[0],concave[0],bottom_type,bottom,no_bevel_b);
    t_type = es_set_type(bevel[1],fillet[1],concave[1],top_type,top,no_bevel_t);
    c_type = es_set_type(bevel[2],fillet[2],concave[2],corner_type,corners);

    // set up feature sizes according to types
    // and limits

    minxyz = min(b_size[0],t_size[0],
                b_size[1],t_size[1],
                z_size) / 2;

    bh = bottom > 0
        && ! es_is_straight(bottom,b_type) ?
            ( limit_size ? min(bottom,minxyz) :         bottom ) : 0;
    th = top > 0 && ! es_is_straight(top,t_type) ?
        ( limit_size ? min(top,minxyz) : top ) : 0;
    ch = corners > 0
        && ! es_is_straight(corners,c_type) ?
            ( limit_size ? min(corners,minxyz) :
                corners ) : 0;

    // for drawing the basic box, treat concave as if
    // it were beveled IF the angle is < 45; this
    // will prevent "debris" from being left over
    // when the edge is cut by the cylinder.

    base_b_type = es_is_concave(bh,b_type) ?
        ( max_b_angle > 135 ? es_none : es_bevel )
        : b_type;
    base_t_type = es_is_concave(th,t_type) ?
        ( max_t_angle > 135 ? es_none : es_bevel )
        : t_type;
    base_c_type = es_is_concave(ch,c_type) ?
        es_bevel : c_type;

    // calculate the basic offsets (ie, for 90° sides)

    b_offset = es_get_offset(bh,base_b_type,ch,base_c_type);
    t_offset = es_get_offset(th,base_t_type,ch,base_c_type);

    base_bh = es_is_straight(bh,base_b_type) ?
        0 : bh;
    base_th = es_is_straight(th,base_t_type) ?
        0 : th;

    bz_offset = base_bh == 0 ? es_min_size : base_bh;

    tz_offset = base_th == 0 ? es_min_size : base_th;

    // calculate the actual coordinates to use

    b_corners = [ [ 0, 0 ], [ b_size[0], b_size[1] ] ];
    t_corners = slope_to_origin ?
        [ [ 0, 0 ], [ t_size[0], t_size[1] ] ]
         : [ [ dx/2, dy/2 ],
            [ t_size[0] + dx/2, t_size[1] + dy/2 ] ];

    b_coords = es_get_coords(b_corners,0,
                    base_b_type,bz_offset,
                    angles[0],b_offset);
    t_coords = es_get_coords(t_corners,z_size,
                    base_t_type,tz_offset,
                    angles[1],t_offset);

    // calculate the offsets needed to center
    // the cube; start by calculating the proper
    // "footprint" to use, based on whether the
    // slope is to origin or to center

    cx = slope_to_origin ? max(t_size[0], b_size[0])
        : b_size[0];
    cy = slope_to_origin ? max(t_size[1], b_size[1])
        : b_size[1];

    c_xyz = [ center || center_x ||
                center_xy || center_xyz.x,
              center || center_y ||
                center_xy || center_xyz.y,
              center || center_z || center_xyz.z ];

    xt = c_xyz.x ? -cx / 2 : 0;
    yt = c_xyz.y ? -cy / 2 : 0;
    zt = c_xyz.z ? -z_size / 2 : 0;

    translate([xt,yt,zt])
    es_difference_if ( enable_difference ) {
        es_hull_if ( enable_hull ) {
            for ( i=[0:3] ) {
                translate(b_coords[i])
                es_make_corner(bh,base_b_type,
                    is_angled,
                    ch,base_c_type,true);
            }
            
            for ( i=[0:3] ) {
                translate(t_coords[i])
                es_make_corner(th,base_t_type,
                    is_angled,
                    ch,base_c_type,false);
            }
        }

        es_make_concave(angles,z_size,
                           b_corners,bh,b_type,
                           t_corners,th,t_type,
                           ch,c_type);
    }
}

module es_make_concave ( angles, z_size,
    b_corners, bh, b_type,
    t_corners, th, t_type,
    ch, c_type )
{
    // use es_get_coords to get top and bottom
    // corner coordinates. For es_get_coords,
    // xy_offset and z_size parameters default to 0;
    // setting z_override to true forces true top/bottom

    bottom_coords_with_offset =
        es_get_coords(b_corners,
            tb_type = b_type, tbh = bh,
            angles = angles[0], z_override = true);
    top_coords_with_offset = 
        es_get_coords(t_corners, z_size = z_size,
            tb_type = t_type, tbh = th,
            angles = angles[1], z_override = true);

    // handle top and bottom
    
    for ( t = [ [ th, t_type, top_coords_with_offset ],
               [ bh, b_type,
                    bottom_coords_with_offset ] ] ) {

        // first parameter [0] = height of feature
        // second parameter [1] = type vector
        // third parameter [2] = corner coordinates

        if ( es_is_concave(t[0],t[1]) ) {
            coords = t[2];

            // handle the primary edges

            for ( i = [0:3] ) {
                start = coords[i];
                end = coords[(i+1) % 4];
                hx = abs(start[0]-end[0]);
                hy = abs(start[1]-end[1]);
                h = max(hx,hy) + es_min_size * 2;
                dx = hx > 0 ? es_min_size : 0;
                dy = hy > 0 ? es_min_size : 0;
                x = min(start[0],end[0]) - dx;
                y = min(start[1],end[1]) - dy;
                z = start[2];

                rot_x = hx > 0 ? 0 : -90;
                rot_y = hy > 0 ? 0 : 90;

                translate([x,y,z])
                rotate([rot_x,rot_y,0])
                cylinder(r=t[0],h=h);
            }

            // handle top/bottom corners when
            // corners are bevel or fillet

            xo = [ 1, -1, -1, 1 ];
            yo = [ 1, 1, -1, -1 ];

            if ( c_type[0] ) {          // corner bevel?
                h = 3*max(ch,t[0]);
                
                for ( i = [0:3] ) {
                    x = coords[i][0] + xo[i] * ch / 2;
                    y = coords[i][1] + yo[i] * ch / 2;
                    z = coords[i][2];
                    rot_z = 90*i + 45;

                    translate([x,y,z])
                    rotate([90,0,rot_z])
                    cylinder(r=t[0],h=h,center=true);
                }

            } else if ( c_type[1] ) {   // corner fillet?

                for ( i = [0:3] ) {
                    x = coords[i][0] + xo[i] * ch;
                    y = coords[i][1] + yo[i] * ch;
                    z = coords[i][2];
                    rot_z = 90*i;

                    translate([x,y,z])
                    rotate([0,0,rot_z])
                    es_make_quarter_toroid(ch,t[0]);
                }
            }
        }
    }
    
    // handle concave corners (includes whatever is
    // needed to match top and bottom)

    if ( es_is_concave(ch,c_type) ) {
        bottom_coords = es_get_coords(b_corners,0,
            angles=angles[0]);
        top_coords = es_get_coords(t_corners,z_size,
            angles=angles[1]);

        for ( i = [0:3] ) {
            top_corner = top_coords[i];
            bottom_corner = bottom_coords[i];
            
            bx = bottom_corner[0];
            by = bottom_corner[1];
            tx = top_corner[0];
            ty = top_corner[1];
            z = top_corner[2];

            translate([bx,by,0])
            skew_over_z(x=tx-bx, y=ty-by, z=z)
            eased_cylinder_cutout ( r=ch, h=z,
                top=th, top_type=t_type,
                bottom=bh, bottom_type=b_type);
        }
    }
}

module es_make_quarter_toroid ( r, h )
{
    clip = r + h + es_min_size;
    
    difference() {
        es_make_toroid ( r, h );
        
        translate([0,-clip,-clip])
        cube([clip*2,clip*2,clip*2]);
        
        translate([-clip,0,-clip])
        cube([clip*2,clip*2,clip*2]);
    }
}

module es_make_corner (tbh,tb_type,is_angled,
    ch,c_type,is_bottom)
{
    clip = ! is_angled;

    if ( es_is_fillet(ch,c_type) ) {
        if ( es_is_fillet(tbh,tb_type) ) {
            es_make_sphere(ch,tbh,is_bottom,
            clipped=clip);
        } else if ( es_is_bevel(tbh,tb_type) ) {
            es_make_cone(ch,tbh,is_bottom);
        } else {
            cylinder(r=ch,h=es_min_size);
        }
    } else {
        if ( es_is_fillet(tbh,tb_type) ) {
            es_make_cylinder(ch,tbh,is_bottom,clip);
        } else if ( es_is_bevel(ch,c_type) ||
                    es_is_bevel(tbh,tb_type) ) {
            es_make_pyramid(ch,tbh,is_bottom);
        } else {
            translate([-es_min_size,-es_min_size,0])
            cube([es_min_size*2,es_min_size*2,es_min_size]);
        }
    }
}

module es_make_sphere ( r, h, is_bottom,
            beak = false, angle = 120, clipped = true )
{
    clip = max(r,h) * 2 + 2;
    z = is_bottom ? h : 0;
    rot = is_bottom ? 180 : 0;

    translate([0,0,z])
    rotate([rot,0,0])
    union() {
        difference() {
            if ( r == h ) {
                sphere(r);
            } else if ( h < r ) {
                es_make_sphere_flattened(r,h);
            } else {
                es_make_sphere_squeezed(r,h);
            }
        
            if ( clipped ) {
                translate([-clip/2,-clip/2,-clip])
                    cube([clip,clip,clip]);
            }
        }
        
        if ( beak && r >= h ) {
            es_make_beak(r-h,h,angle,true,true);
        }
    }
}

module es_make_sphere_flattened ( r, h )
{
    if ( r == h ) {
        sphere(r);
    
    } else if ( h > r ) {
        rotate([90,0,0])
        es_make_sphere_flattened ( h, r );

    } else {
        off = r-h;
        clip = h*2+2;

        rotate_extrude()
        difference() {
            union() {
                translate([0,-h]) square([off,h*2]);
                translate([off,0,0]) circle(h);
            }

            translate([-clip,-clip/2]) square(clip);
        }
    }
}

module es_make_sphere_squeezed ( r, h )
{
    k = 2 * sqrt(2) - 2;
    extra = k * ( h - r );
    
    for ( rot = [ 45, 135 ] ) {
        rotate([0,0,rot])
        es_make_oblong(r,h,extra);
    }
}  

module es_make_oblong ( r, h, extra )
{
    union() {
        for ( x = [ extra/2, -extra/2 ] ) {
            translate([x,0,0])
            es_make_sphere_flattened(r,h);
        }

        rotate([0,90,0])
        linear_extrude(height=extra,center=true)
        projection()
        rotate([0,90,0])
        es_make_sphere_flattened(r,h);
    }
}

module es_make_cone ( r, h, is_bottom )
{
    if ( h > r ) {
        flip = is_bottom ? 180 : 0;
        trans = is_bottom ? h : 0;

        translate([0,0,trans])
        rotate([flip,0,0])
        es_make_cone_slanted(r,h);

//      To change to the alternate corner shape, comment
//      out the previous line and uncomment the following:
//      es_make_cone_pyramid(r,h);
        
    } else { 
        r2 = r-h;

        if ( is_bottom ) {
            cylinder(r1=r2,r2=r,h=h);
        } else {
            cylinder(r1=r,r2=r2,h=h);
        }
    }
}

module es_make_cone_slanted ( r, h )
{
    trans = r - h;

    for ( rot = [0:90:270] ) {
        rotate([0,0,rot])
        linear_extrude(height=h,scale=0)
        translate([trans,trans,0])
        circle(r);
    }
}        

module es_make_cone_pyramid ( r, h )
{
    upper = h - r;

    translate([-upper,-upper,r-es_min_size])
    cube([upper*2,upper*2,es_min_size]);

    translate([0,0,r])
    linear_extrude(height=upper,scale=0)
    square(upper*2,center=true);
        
    for ( rot = [0:90:270] ) {
        rotate([0,0,rot])
        translate([upper,upper,0])
        cylinder(r1=r,r2=0,h=r);
    }
}

module es_make_cylinder ( r, h, is_bottom,
    clipped = true )
{
    clip = ( r + h ) * 2 + .2;
    z = is_bottom ? h : 0;
    rot = is_bottom ? 180 : 0;
    r2 = sqrt(2) * ( h + r / 2 ) - h;

    translate([0,0,z])
    rotate([0,rot,0])
    difference() {
        intersection() {
            es_make_cylinder_flattened(r,h);
        
            if ( r > 0 ) {
                rotate([0,0,45])
                es_make_cylinder_flattened(r2,h);
            }
        }        

        if ( clipped ) {
            translate([0,0,-clip/2])
            cube([clip,clip,clip],center=true);
        }
    }
}

module es_make_cylinder_flattened ( r, h )
{
    extra = r;
    height = ( h + extra ) * 2 + .2;

    intersection_for ( rot = [ 0, 90 ] ) {
        rotate([0,0,rot])

        for ( rot = [0,180] ) {
            rotate([0,0,rot])
            rotate([90,0,0])
            linear_extrude(height=height,center=true)
            difference() {
                union() {
                    translate([0,-h,0])
                    square([extra,h*2]);
                    translate([extra,0,0])
                    circle(r=h);
                }

                translate([-h-1,0,0])
                square(size=[h*2+2,h*2+2],center=true);
            }
        }
    }
}

module es_make_pyramid ( r, h, is_bottom )
{
    if ( h == 0 ) {
        rotate([0,0,45])
        translate([0,0,es_min_size/2])
        cube([r*sqrt(2),r*sqrt(2),es_min_size],center=true);

    } else {
        flip = is_bottom ? 180 : 0;
        trans = is_bottom ? h : 0;

        translate([0,0,trans])
        rotate([flip,0,0])

        if ( r == 0 ) {
            linear_extrude(h,scale=0)
            square(size=h*2,center=true);

        } else if ( r < h ) {
            for ( rot = [0:90:270] ) {
                rotate([0,0,rot])
                linear_extrude(h,scale=0)
                translate([r-h,r-h,0])
                rotate([0,0,45])
                square(size=r*sqrt(2),center=true);
            }

        } else {
            rotate([0,0,45])
            linear_extrude(h,scale=(r-h)/r)
            square(size=r*sqrt(2),center=true);
        }
    }
}

