// eased_shape cutout modules
// filename: es_cutout.scad

/* Copyright 2018-2020 Andrew H. Wakefield

    es_cutout.scad is part of the eased_shapes.scad library

    This library is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

module eased_cylinder_cutout (
        // basic size parameters

        r = -1,         // radius
        r1 = -1,        // bottom radius (overrides r)
        r2 = -1,        // top radius (overrides r)

        h = 2,         // height

        top = 0,       // size of top feature
        bottom = 0,    // size of bottom feature
        corners = 0,   // size of half-cylinder edge

        // vectors to enable feature type
        // for [top,bottom,half-cylinder edges]

        bevel = es_none,
        fillet = es_none,     // outward fillet
        concave = es_none,    // inward fillet
        beak = es_none,

        // vectors to set feature type as
        // [bevel,fillet,concave,beak]

        top_type = es_none,     // type of top feature
        bottom_type = es_none,  // type of bottom feat.
        corner_type = es_none,  // type of half-cyl. edge

        // if enabled above, a corner or "beak"
        // is added to x-axis side using the angle(s)
        // specified below; beak angle is valid only   
        // if > 0 and < 180
        
        beak_angle = 120,   // default angle of "beak"

        // if > 0, these override the beak angle
        // at the top / bottom

        top_beak_angle = -1,
        bottom_beak_angle = -1,

        // if arc > 0, cut the cylinder into a wedge
        // of angle = arc

        arc = 0,

        // if true, center the cylinder

        center = false,

        // if half_cylinder = true, include cutout to
        // shape edges of a half-cylinder cutout rotated
        // to the specified angle
        
        half_cylinder = false,
        half_cylinder_rot = 0,

        // if false, only show ends
  
        enable_middle = true
    )
{
    br = r1 >= 0 ? r1 : max(0,r);
    tr = r2 >= 0 ? r2 : max(0,r);

    if ( br + tr <= 0 || h <= 0 )
        echo("Invalid parameters!");

    // Note that fillet and concave are swapped because
    // one is the cutout shape of the other

    b_type = es_set_type( bevel[0], concave[0], fillet[0],
        [ bottom_type[0], bottom_type[2],
          bottom_type[1], bottom_type[3] ],
          bottom, beak = beak[0]) ;

    t_type = es_set_type( bevel[1], concave[1], fillet[1],
        [ top_type[0], top_type[2], 
          top_type[1], top_type[3] ],
          top, beak = beak[1]);

    ct_type = es_set_type( bevel[0], fillet[0], concave[0],
      top_type, top );

    cb_type = es_set_type( bevel[1], fillet[1], concave[0],
      bottom_type, bottom );

    c_type = es_set_type( bevel[2], concave[2], fillet[2],
        [ corner_type[0], corner_type[2], 
          corner_type[1], corner_type[3] ],
          corners );

    th = top > 0 && ! es_is_straight(top,t_type) ?
        top : 0;
    bh = bottom > 0 && ! es_is_straight(bottom,b_type) ?
        bottom : 0;
    ch = corners > 0 && ! es_is_straight(corners,c_type) ?
        corners : 0;

    ba = max(es_min_size,
            min(180-es_min_size,beak_angle));

    tba = top_beak_angle > 0 
        && top_beak_angle< 180 ? 
         top_beak_angle :
          ba;
    bba = bottom_beak_angle > 0
        && bottom_ba < 180 ?
         bottom_beak_angle :
          ba;

    slope = ( br - tr ) / h;

    zt = center ? -h/2 : 0;

    translate([0,0,zt])
    union() {
        if ( enable_middle ) {
            translate([0,0,-es_min_size])
            eased_cylinder(r1=br+es_min_size*slope,
                r2=tr-es_min_size*slope,
                h=h+es_min_size*2,
                beak=[t_type[3],b_type[3]],
                top_beak_angle = tba,
                bottom_beak_angle = bba,
                arc = arc);
        }

        if ( ! es_is_straight(bh,b_type) ) {
            r = (br+bh)-bh*slope;

            translate([0,0,-es_min_size])
            eased_cylinder(
                r=r,
                top=bh, bottom=0,
                h=bh+es_min_size,
                bevel=[false,false],
                top_type=b_type,
                beak=[b_type[3],b_type[3]],
                top_beak_angle=bba,
                bottom_beak_angle=bba,
                arc = arc
            );
        }

        if ( ! es_is_straight(th,t_type) ) {
            r = (tr+th)+th*slope;

            translate([0,0,h-th])
            eased_cylinder(
                r=r,
                bottom=th, top=0,
                h=th+es_min_size,
                bevel=[false,false],
                bottom_type=t_type,
                beak=[t_type[3],t_type[3]],
                top_beak_angle=tba,
                bottom_beak_angle=tba,
                arc = arc
            );
        }
        
        if ( half_cylinder &&
              ! es_is_straight(ch,c_type) ) {
          rotate([0,0,half_cylinder_rot])
          translate([-ch,0,0])
          eased_cube_cutout(
            size_bottom = [br,br*2+ch*2,h],
            size_top = [tr,tr*2+ch*2,h],
            top = th, top_type = ct_type,
            bottom = bh, bottom_type = cb_type,
            corners = ch, corner_type = c_type,
            center_y = true
          );
        }          
    }            
}

module eased_cube_cutout (


        // set dimensions, with several options. The
        // actual top and bottom sizes are determined
        // by the highest-priority valid setting (=>0)
        // from size, size_top, and size_bottom; valid
        // dimensions from size_top and size_bottom
        // override corresponding dimensions in size.
        // For z dimension, height has the highest
        // priority, followed by h, then size_top[2],
        // size_bottom[2], and size[2]. size, size_top,
        // and size_bottom are vectors containing
        // [x, y, z]; height and h are scalar.

        size = [-1,-1,-1],
        size_bottom = [-1,-1,-1],
        size_top = [-1,-1,-1],
        h = -1,                   // alternate z dimension
        height = -1,              // another alternate z

        // size of top, bottom, and corner features

        bottom = 0,
        top = 0,
        corners = 0,
                                  
        // bevel, fillet, concave consist of a vector of
        // true/false values for [top,bottom,corners],
        // indicating whether this type of feature is
        // to be used. Note that concave overrides
        // fillet overrides bevel.

        bevel = es_none,
        fillet = es_none,         // outward fillet
        concave = es_none,      // inward fillet

        // top_type, bottom_type, corner_type consist of
        // a vector of true/false values for [bevel, fillet,
        // concave], indicating which type of feature
        // is applied to the top, bottom, or corners.
        // Again, concave overrides fillet overrides bevel
 
        bottom_type = es_none,
        top_type = es_none,
        corner_type = es_none,

        // slope_to_origin applies only when top and
        // bottom sizes differ. If slope_to_origin is true,
        // any slope will occur on the right and back
        // sides; if false any slope will occur equally
        // on left/right and on front/back
        
        slope_to_origin = false,
 
        // if any parameter below is true, the
        // cube is centered on the specified axis/axes.
        // Note that the centering is based on the largest
        // face, top or bottom, including where it winds
        // up as a result of slope_to_origin setting

        center = false,
        center_x = false,
        center_y = false,
        center_z = false,
        center_xy = false,
        center_xyz = [false, false, false],

        // natural_corners controls the treatment of
        // fillet and concave corners. If false, setting
        // a fillet creates a concave corner (giving a
        // filleted cutout) and vice-versa. If true, this
        // action is inverted
        
        natural_corners = true,

        // debugging settings; for normal operation, leave
        // these set to true. If enable_middle is false,
        // only the top and bottom extensions are shown,
        // without the rest of the cube in the middle.
        // If enable difference is false, the corner cutout
        // shapes will be shown rather than cut from
        // the top and bottom extensions.

        enable_middle = true,
        enable_difference = true
    )
{
    // height has highest priority for z, followed by
    // h, size_top, size_bottom, then size;
    // if none of these is valid, default=10

    z_size = es_get_priority_parameter (
        [height,h,size_top[2],size_bottom[2],size[2],10] );

    // size_top and size_bottom dimensions for x, y
    // have highest priority, followed by size;
    // if none of these is valid, default=10.
    // z is set to 0 for b_size and to z_size for t_size

    b_size = es_assemble_priority_vector (
        [ [ size_bottom[0], size_bottom[1], 0 ],
          [ size[0], size[1], 0 ], [10,10,0] ] );
    t_size = es_assemble_priority_vector (
        [ [-1,-1,z_size], size_top, size, [10,10,10] ] );

    dx = b_size[0] - t_size[0];
    dy = b_size[1] - t_size[1];

    // calculate the angles for each side

    bx_angle = slope_to_origin ?
        es_get_angle(z_size,dx) :
        es_get_angle(z_size,dx/2);

    by_angle = slope_to_origin ?
        es_get_angle(z_size,dy) :
        es_get_angle(z_size,dy/2);

    is_angled = ( bx_angle != 90 || by_angle != 90 );
    
    angles = // bottom: [ left, right, front, back ]
             // top : [ left, right, front, back ]
        slope_to_origin ?
            [ [ 90, bx_angle,
                90, by_angle ],
              [ 90, 180-bx_angle,
                90, 180-by_angle ] ]
          : [ [ bx_angle, bx_angle,
               by_angle, by_angle ],
              [ 180-bx_angle, 180-bx_angle,
                180-by_angle, 180-by_angle ] ];

    min_b_angle = min(bx_angle,by_angle);
    min_t_angle = min(180-bx_angle,180-by_angle);

    no_bevel_b = min_b_angle <= 45;
    no_bevel_t = min_t_angle <= 45;

    // for top and bottom features, note that fillet and
    // concave are switched to make the appropriate
    // cutout shape (each is the cutout of the other).

    b_type = es_set_type(bevel[0],concave[0],fillet[0],
       [ bottom_type[0], bottom_type[2],
         bottom_type[1] ], bottom, no_bevel_b);
    t_type = es_set_type(bevel[1],concave[1],fillet[1],
       [ top_type[0], top_type[2], top_type[1] ], top,
       no_bevel_t);

    // the default "natural" corners indicate the actual
    // shape of the corners on the cutout cube. If set
    // to false, fillet and concave are inverted as above

    c_type = natural_corners ?
        es_set_type(bevel[2],fillet[2],concave[2],
            corner_type,corners) :
        es_set_type(bevel[2],concave[2],fillet[2],
            [ corner_type[0], corner_type[2],
                corner_type[1] ], corners);

    // set up feature sizes

    th = top > 0 && ! es_is_straight(top,t_type) ?
        top : 0;
    bh = bottom > 0 && ! es_is_straight(bottom,b_type) ?
        bottom : 0;
    ch = corners > 0 && ! es_is_straight(corners,c_type) ?
        corners : 0;

    // calculate the coordinates to use for the top
    // and bottom cubes, using -bh / -th as the
    // xy_offset in order to allow for the larger size
    // of the extension. bh / th is useds as the z_offset
    // to calculate where the extension will meet the
    // base cube (allowing for the slope indicated
    // by the angles parameter at bh/th distance)

    b_corners = [ [ 0, 0 ], [ b_size[0], b_size[1] ] ];
    t_corners = slope_to_origin ?
        [ [ 0, 0 ], [ t_size[0], t_size[1] ] ]
         : [ [ dx/2, dy/2 ],
            [ t_size[0] + dx/2, t_size[1] + dy/2 ] ];

    b_coords = es_get_coords(corners = b_corners,
        z_size = 0, tb_type = es_none, tbh = bh,
        angles = angles[0], xy_offset = -bh);
    t_coords = es_get_coords(corners = t_corners,
        z_size = z_size, tb_type = es_none, tbh = th,
        angles = angles[1], xy_offset = -th);

    // origin of top and bottom extension cubes
    
    bx = b_coords[0][0];
    by = b_coords[0][1];

    tx = t_coords[0][0];
    ty = t_coords[0][1];
    tz = t_coords[0][2];

    // size of top and bottom extension cubes

    bx_size = b_coords[1][0] - bx;
    by_size = b_coords[2][1] - by;

    tx_size = t_coords[1][0] - tx;
    ty_size = t_coords[2][1] - ty;

    // offset to achieve center for x, y

    cx = tx_size > bx_size ? tx + tx_size / 2
        : bx + bx_size / 2;
    cy = ty_size > by_size ? ty + ty_size / 2
        : by + by_size / 2;

    // translation to be applied

    c_xyz = [ center || center_x ||
                center_xy || center_xyz.x,
              center || center_y ||
                center_xy || center_xyz.y,
              center || center_z || center_xyz.z ];

    xt = c_xyz.x ? -cx : 0;
    yt = c_xyz.y ? -cy : 0;
    zt = c_xyz.z ? -z_size / 2 : 0;

    // corner size to use for the extensions

    bch = es_get_coc(bh,b_type,ch,c_type);
    tch = es_get_coc(th,t_type,ch,c_type);

    translate([xt,yt,zt])
    union() {
        if ( enable_middle ) {
            translate([0,0,-es_min_size])
            eased_cube ( size_bottom = b_size,
                size_top = t_size,
                height = z_size + es_min_size * 2,
                top = 0, bottom = 0, corners = ch,
                corner_type = c_type,
                slope_to_origin = slope_to_origin,
                limit_size = false);
        }

        if ( bh > 0 ) {
            size = [ bx_size, by_size, bh+es_min_size ];

            es_difference_if(enable_difference) {
                translate([bx,by,-es_min_size])
                eased_cube(size,
                    top=bh,bottom=0,corners=bch,
                    top_type=b_type,
                    corner_type=c_type,
                    limit_size=false);

                if ( es_is_cc(bh,b_type,ch,c_type) ) {
                    es_make_cc(b_coords,bh,ch,true);
                } else
                  if ( es_is_cb(bh,b_type,ch,c_type) ) {
                    es_make_cb(b_coords,bh,ch,true);
                }
            }
        }

        if ( th > 0 ) {
            size = [ tx_size, ty_size, th+es_min_size ];

            es_difference_if(enable_difference) {
                translate([tx,ty,tz])
                eased_cube(size,
                    top=0,bottom=th,corners=tch,
                    bottom_type=t_type,
                    corner_type=c_type,
                    limit_size=false);

                if ( es_is_cc(th,t_type,ch,c_type) ) {
                    es_make_cc(t_coords,th,ch,false);
                } else 
                  if ( es_is_cb(th,t_type,ch,c_type) ) {
                    es_make_cb(t_coords,th,ch,false);
                }
            }
        }
    }
}

module es_make_cc ( coords, tbh, ch, is_bottom )
{
    z = is_bottom ? tbh : coords[0][2];

    rot = [ 45, 135, 225, 315 ];
    dx = [ 1, -1, -1, 1 ];
    dy = [ 1, 1, -1, -1 ];

    for ( i = [0:3] ) {
        x = coords[i][0] + tbh * dx[i];
        y = coords[i][1] + tbh * dy[i];
        
        translate([x,y,z])
        rotate([0,0,rot[i]])
        translate([-max(tbh,ch),0,0])
        es_make_oblong(ch,tbh,ch*2);
    }
}

module es_make_cb ( coords, tbh, ch, is_bottom )
{
    z = is_bottom ? 0 : coords[0][2];

    rot = [ 0, 90, 180, 270 ];
    dx = [ 1, -1, -1, 1 ];
    dy = [ 1, 1, -1, -1 ];

    for ( i = [0:3] ) {
        x = coords[i][0] + tbh * dx[i];
        y = coords[i][1] + tbh * dy[i];

        translate([x,y,z])
        rotate([0,0,rot[i]])
        es_make_cb_corner(tbh,ch,is_bottom);
    }        
}

module es_make_cb_corner ( tbh, ch, is_bottom )
{
    h = es_min_size / 10;

    rot = is_bottom ? 180 : 0;
    x = is_bottom ? tbh : tbh;
    y = is_bottom ? -tbh : tbh;
    z = is_bottom ? tbh : 0;
    
    translate([-tbh,-tbh,z])
    rotate([rot,0,0])
    linear_extrude(height=tbh,scale=0)
    translate([x,y,0])
    circle(r=ch);
    
    translate([0,0,z-h])
    cylinder(r=ch,h=h*2);
}
