// eased shape cylinder modules
// filename: es_cylinder.scad

/* Copyright 2018-2020 Andrew H. Wakefield

    es_cylinder.scad is part of the eased_shapes.scad library

    This library is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

module eased_cylinder (

        // basic size parameters

        r = -1,         // radius
        r1 = -1,        // bottom radius (overrides r)
        r2 = -1,        // top radius (overrides r)

        h = 2,         // height

        top = 0,       // size of top feature
        bottom = 0,   // size of bottom feature

        // vectors to enable feature type for [top,bottom]

        bevel = es_none,
        fillet = es_none,     // outward fillet
        concave = es_none,  // inward fillet
        beak = es_none,

        // vectors to set feature type as
        // [bevel,fillet,concave,beak]

        top_type = es_none,     // type of top feature
        bottom_type = es_none, // type of bottom feat.

        // if enabled above, a corner or "beak"
        // is added to x-axis side using the angle(s)
        // specified below; beak angle is valid only   
        // if > 0 and < 180
        
        beak_angle = 120,   // default angle of "beak"

        // if > 0, these override the beak angle
        // at the top / bottom

        top_beak_angle = -1,
        bottom_beak_angle = -1,

        // if arc > 0, cut the cylinder into a wedge
        // of angle = arc

        arc = 0,

        center = false,            // center the cylinder

        limit_size = true,         // if true, limit edge
                                   // size to <= radius
        enable_hull = true,       // if false, show
                                   // ends only
        enable_difference = true  // if false, show
                                   //toroids
    )
{
    br = es_get_priority_parameter([r1,r,10]);
    tr = es_get_priority_parameter([r2,r,10]);

    if ( br + tr <= 0 || h <= 0 )
        echo("Invalid parameters!");

    bottom_angle = es_get_angle(rise=h,run=br-tr);
    top_angle = 180 - bottom_angle;

    no_bevel_b = bottom_angle >= 135;
    no_bevel_t = top_angle >= 135;

    b_type = es_set_type(bevel[0],fillet[0],concave[0],bottom_type,bottom,no_bevel_b,beak[0]);
    t_type = es_set_type(bevel[1],fillet[1],concave[1],top_type,top,no_bevel_t,beak[1]);

    bh = bottom > 0 && ! es_is_straight(bottom,b_type) ?
        ( limit_size ? min(bottom,h,br*2) : bottom ) : 0;
    th = top > 0 && ! es_is_straight(top,t_type) ?
        ( limit_size ? min(top,h,tr*2) : top ) : 0;

    // beak angle limited to 0 > ba > 180;

    ba = max(es_min_size,
        min(180-es_min_size,beak_angle));

    tba = top_beak_angle > 0
        && top_beak_angle < 180 ?
         top_beak_angle : ba;
    bba = bottom_beak_angle > 0
        && bottom_beak_angle < 180 ?
         bottom_beak_angle : ba;

    gamma = arc > 0 ? ( arc % 360 ) / 2 : 0;
    
    zt = center ? -h/2 : 0;

    // slope is inverted ( run / rise ) to avoid
    // division by 0 ( if br = tr ).

    inv_slope = ( br - tr ) / h;

    translate([0,0,zt])
    es_difference_if ( enable_difference ) {
        es_hull_if ( enable_hull ) {
            es_make_end(br,bh,bottom_angle,b_type,
                true,bba);
            translate([0,0,h])
            es_make_end(tr,th,top_angle,t_type,
                false,tba);
        }
     
        if ( b_type[2] ) {
            angle = bottom_angle;
            cx = bh * cos(angle/2) / sin(angle/2);
            r = br - ( cx - bh );
            es_make_toroid(r,bh,
                beak=b_type[3],angle=bba);
        }
        
        if ( t_type[2] ) {
            angle = top_angle;
            cx = th * cos(angle/2) / sin(angle/2);
            r = tr - ( cx - th );
            translate([0,0,h])
            es_make_toroid(r,th,
                beak=t_type[3],angle=tba);
        }

        if ( gamma > 0 && gamma < 180 ) {
            es_make_arc_cutout(max(br,tr), h, gamma);
        }
    }
}

module es_make_arc_cutout ( r, h, gamma )
{
    r1 = r + es_min_size;
    height = h + 2 * es_min_size;

    x = r1 * cos(gamma);
    y = r1 * sin(gamma);
    
    p = ( x <= 0 ) ?
          [ [ x, y ], [ 0, 0 ], [ x, -y ],
            [ -r1, -y ], [ -r1, y ], [ x, y ] ]
        : [ [ x, y ], [ 0, 0 ], [ x, -y ],
           [ x, -r1], [ -r1, -r1 ],
           [ -r1, r1 ] , [ x, r1 ], [ x, y ] ];

    translate([0,0,-es_min_size])
    linear_extrude(height=height)
    polygon(p);
}

module es_make_end ( r, h, angle, type, is_bottom, ba )
{
    height = type[2] && angle > 45 ? es_min_size
        : max(es_min_size,h);

    if ( type[1] ) {    // filleted end
        clipped = ( angle == 90 );

        cx = height * cos(angle/2) / sin(angle/2);
        r1 = r - ( cx - height );
        zt = is_bottom ? 0 : -height;

        translate([0,0,zt])
        es_make_sphere(r1,height,is_bottom,
            beak=type[3],angle=ba,clipped=clipped);

    } else {           // all others
        rot = is_bottom ? 0 : 180;

        r2 = ( angle == 90 ) ? r
            : r - 1/tan(angle) * height;
        r1 = es_is_straight(height,type) ? r : r2 - height;

        rotate([rot,0,0])
        union() {
            cylinder(r1=r1,r2=r2,h=height);

            if ( type[3] ) {
                gamma = ( 180 - ba ) / 2;
                x1 = cos(gamma) * r1;
                y1 = sin(gamma) * r1;
                x2 = y1 * tan(gamma);

                translate([x1,0,0])
                skew_over_z(x=(r2-r1)*cos(gamma),
                    z=height)
                linear_extrude(height=height,scale=r2/r1)
                polygon([[0,-y1],[0,y1],
                     [x2,0],[0,-y1]]);
            }
        }
    }
}

module es_make_toroid ( r, h, 
            beak = false, angle = 120 )
{
    gamma = (180-angle) / 2;
    cg = cos(gamma);
    sg = sin(gamma);

    clip = max(r,h) * 4 + 2;

    difference() {
        rotate_extrude()
        difference() {
            translate([r,0,0]) circle(h);
            translate([-clip*2,-clip,0]) 
            square(clip*2);
        }

        if ( beak ) {
            x = (r + h) * 2 * cg;
            y = (r + h) * 2 * sg;

            linear_extrude(
                height=(h+es_min_size)*2,
                center=true)
            polygon( [ [0,0], [x,-y], [x,y], [0,0] ] );
        }
    }

    if ( beak ) {
        es_make_beak(r,h,angle,false);
    }
}

module es_make_beak ( r, h, angle,
            fill = false, clip = false )
{
    gamma = (180-angle) / 2;
    cg = cos(gamma);
    sg = sin(gamma);

    chord = 2 * ( r+h ) * sg;
    length = chord / ( 2 * cg ) + 2 * es_min_size;

    x = r * cg;
    y = r * sg;

    l = x + length + 2 * h + 1;
    w = r + h + 1;

    difference() {
        union() {
            for ( v = [ -1, 1 ] ) {
                difference() {
                    translate([x,v*y,0])
                    rotate([0,0,v*(gamma-90)])
                    translate([-es_min_size,0,0])
                    rotate([0,90,0])
                    cylinder(r=h,h=length);

                    translate([l/2-h-1,
                        -v*(w/2+es_min_size),0])
                    cube([l, w, h*2 + 2], center=true);
                }                
            }
    
            if ( fill ) {
                x2 = r / cg;

                translate([0,0,-h])
                linear_extrude(height=2*h)
                polygon( [ [ x, y ], [ x, -y ],
                        [ x2, 0 ], [ x, y ] ] );
            }
        }
        
        if ( clip ) {
            x3 = w / cg;

            translate([-w,-w,-h-1])
            cube([w+x3,w*2,h+1]);
        }
    }
}    

