// eased_shape functions
// filename: es_functions.scad

/* Copyright 2018-2020 Andrew H. Wakefield

    es_functions.scad is part of the eased_shapes.scad library

    This library is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

function es_is_odd ( n )
    = ( floor ( n / 2 ) * 2 != n );

function es_is_even ( n )
    = ( floor ( n / 2 ) * 2 == n );

function es_get_angle ( rise, run )
    = run == 0 ? 90
        : rise / run < 0 ?
          180 - atan( abs( rise / run ) )
           : atan( rise / run );

// recursive function to return the highest priority
// valid member (>= 0) of the params vector. To
// avoid indeterminate results, include a valid
// default as the last member of params

function es_get_priority_parameter ( params, index=0 )
    =   index == len(params) - 1 || 
            params[index] >= 0 ? params[index]
            : es_get_priority_parameter(params,index+1);

// recursive function to return a vector representing
// a specific column in a vector-of-vectors (vectors)

function es_extract_column ( vectors, item, index=0 )
    =   index == len(vectors) - 1
            ? [ vectors[index][item] ]
            : concat( [ vectors[index][item] ],
              es_extract_column(vectors, item, index+1) );

// recursive function to extract the highest-priority
// parameters from a vector-of-vectors (vectors)

function es_assemble_priority_vector ( vectors,
    item = 0 )
    =   item == len(vectors[0]) - 1
            ? [ es_get_priority_parameter ( 
                    es_extract_column(vectors,item) ) ]
            : concat( [ es_get_priority_parameter ( 
                es_extract_column(vectors,item) ) ],
                    es_assemble_priority_vector ( 
                        vectors, item+1 ) );

// helper function for the inside_corner module;
// returns a vector representing the direction of
// the corner [ on_x, on_y, on_z ] based on
// either the dir vector or the individual on_?
// parameters

function es_set_dir ( on_x, on_y, on_z, dir )
    =   on_z || dir[2] ? es_on_z
         : on_y || dir[1] ? es_on_y
          : es_on_x;

// helper function for the eased_cube/cylinder
// modules; selects settings from a type vector
// and from individual parameters and returns a
// vector indicating the highest-priority type,
// plus whether or not a beak is included. Return
// vector = [ bevel, fillet, concave, beak ]

function es_set_type ( bevel, fillet, concave, type,
            h, no_bevel = false, beak = false )
    =   [ ( bevel || type[0] ) && h > 0
            && ! no_bevel &&
            ! ( concave || fillet || type[1] || type[2] ),
          ( fillet || type[1] ) && h > 0 &&
            ! ( concave || type[2] ),
          ( concave || type[2] ) && h > 0,
          ( beak || type[3] ) ];

// alternate version that takes a vector of vectors,
// each containing bevel / fillet / concave and optionally
// beak; returns a vector with the highest priority
// feature selected ( concave > fillet > bevel ) plus
// optional beak; note that no feature will be selected
// if h <= 0

function es_set_type_v ( h, type_v, no_bevel = false )
    =   let ( bevel = es_vector_or (
                es_extract_column ( type_v, 0 ) ),
              fillet = es_vector_or (
                es_extract_column ( type_v, 1 ) ),
              concave = es_vector_or (
                es_extract_column ( type_v, 2 ) ),
              beak = es_vector_or (
                es_extract_column ( type_v, 3 ) ) )
                
        [   bevel && h > 0 &&
              ! ( no_bevel || fillet || concave ),
            fillet && h > 0 && ! concave,
            concave && h > 0,
            beak ];

// helper function to return the OR value of a
// set of true/false values contained in a vector
// of unknown size; note that a zero-length vector
// is allowed and will return false

function es_vector_or ( v, index = 0 )
    =   index == len(v)
            ? false
            : v[index] || es_vector_or(v,index+1);

// the four es_is_xyz functions below test the type
// and size (h) of a feature to determine whether
// that feature should be applied

function es_is_straight ( h, type )
    = h <= 0 || ! ( type[0] || type[1] || type[2] );

function es_is_bevel ( h, type )
    = type[0] && h > 0;

function es_is_fillet ( h, type )
    = type[1] && h > 0;

function es_is_concave ( h, type )
    = type[2] && h > 0;

// the various es_is_?? functions below test the
// type and size of a top or bottom feature against
// the type and size of a corner feature to determine
// whether a particular combination is to be applied

function es_is_ff ( tbh, tb_type, ch, c_type )
    =   es_is_fillet(ch,c_type)         // is corner fillet?
        && es_is_fillet(tbh,tb_type);   // is top/bottom fillet?

function es_is_fb ( tbh, tb_type, ch, c_type )
    =   es_is_fillet(ch,c_type)        // is corner fillet?
        && es_is_bevel(tbh,tb_type); // is t/b bevel?

function es_is_fs ( tbh, tb_type, ch, c_type )
    =   es_is_fillet(ch,c_type)        // is corner fillet?
        && es_is_straight(tbh,tb_type);

function es_is_bf ( tbh, tb_type, ch, c_type )
    =   es_is_bevel(ch,c_type)          // is corner bevel?
        && es_is_fillet(tbh,tb_type);    // is t/b fillet?

function es_is_bc ( tbh, tb_type, ch, c_type )
    =   es_is_bevel(ch,c_type)           // is corner bevel?
        && es_is_concave(tbh,tb_type);  // is t/b concave?

function es_is_bb ( tbh, tb_type, ch, c_type )
    =   es_is_bevel(ch,c_type)          // is corner bevel?
        && es_is_bevel(tbh,tb_type);    // is t/b fillet?

function es_is_bs ( tbh, tb_type, ch, c_type )
    =   es_is_bevel(ch,c_type)    // is corner bevel?
        && es_is_straight(tbh,tb_type);

function es_is_cb ( tbh, tb_type, ch, c_type )
    =   es_is_concave(ch,c_type)     // is corner concave?
        && es_is_bevel(tbh,tb_type);   // is t/b bevel?

function es_is_cf ( tbh, tb_type, ch, c_type )
    =   es_is_concave(ch,c_type)     // is corner concave?
        && es_is_fillet(tbh,tb_type);  // is t/b fillet?

function es_is_cc ( tbh, tb_type, ch, c_type )
    =   es_is_concave(ch,c_type)       // is corner concave?
        && es_is_concave(tbh,tb_type); // is t/b concave?

function es_is_sf ( tbh, tb_type, ch, c_type )
    =   es_is_fillet(tbh,tb_type)    // is t/b fillet?
        && es_is_straight(ch,c_type);

function es_is_sb ( tbh, tb_type, ch, c_type )
    =   es_is_bevel(tbh,tb_type)   // is t/b bevel?
        && es_is_straight(ch,c_type);

function es_is_ss ( tbh, tb_type, ch, c_type )
    =   es_is_straight(tbh,tb_type)
        && es_is_straight(ch,c_type);

// calculate the offset to apply to a corner block
// based on the type of corner used

function es_get_offset(tbh, tb_type, ch, c_type)
    = es_is_ff(tbh,tb_type,ch,c_type) && tbh > ch ?
     tbh : // offset for "squeezed sphere"
      es_is_fb(tbh,tb_type,ch,c_type) && tbh > ch ?
       tbh : // offset for "slanted cone"
        es_is_fillet(ch,c_type) ?
         ch : // offset for other cones, spheres, disks
          es_is_fillet(tbh,tb_type) ?
           ch+tbh : // offset for cylinder-edge cubes
                    // and octagons
            es_is_ss(tbh,tb_type,ch,c_type) ?       
             es_min_size : // offset for minimum block
              es_is_sb(tbh,tb_type,ch,c_type) ?
               tbh : // offset for basic pyramids 
                es_is_bb(tbh,tb_type,ch,c_type) ?
                 max(ch,tbh) : // offset for 45-degree
                               // or octoganal pyramids
                  ch; // default offset

// calculate the size for a "cut-out corner"

function es_get_coc ( tbh, tb_type, ch, c_type )
    =   ch <= 0 ? 0 :
         es_is_bc(tbh,tb_type,ch,c_type) ||
          es_is_bf(tbh,tb_type,ch,c_type) ?
           ch + 2 * tbh * tan(22.5) / sqrt(2) :
            es_is_cf(tbh,tb_type,ch,c_type) ?
              sqrt( tbh*tbh + (tbh+ch)*(tbh+ch) ) - tbh :
               es_is_concave(ch,c_type) ? 0 :
                ch + tbh;

// assemble the coordinates for a set of top or
// bottom corner blocks

function es_get_coords ( corners, z_size = 0,
    tb_type = es_none, tbh = 0,
    angles = [ 90, 90, 90],
    xy_offset = 0, z_override = false )
  =
    let ( is_bottom = ( z_size == 0 ),
         is_fillet = es_is_fillet(tbh,tb_type),
         x0 = corners[0][0], y0 = corners[0][1],
         x1 = corners[1][0], y1 = corners[1][1],
         z = is_bottom ? 0 : 
            z_override ? z_size :
                z_size - tbh, 
         ldx = ( angles[0] == 90 ) ? 0 :
            is_fillet ? tbh / tan(angles[0]/2) - tbh
                : tbh / tan(angles[0]),
         rdx = ( angles[1] == 90 ) ? 0 :
            is_fillet ? tbh / tan(angles[1]/2) - tbh
                : tbh / tan(angles[1]),
         fdy = ( angles[2] == 90 ) ? 0 :
            is_fillet ? tbh / tan(angles[2]/2) - tbh
                : tbh / tan(angles[2]),
         bdy = ( angles[3] == 90 ) ? 0 :
            is_fillet ? tbh / tan(angles[3]/2) - tbh
                : tbh / tan(angles[3])
       )

    [   [ x0+xy_offset+ldx, y0+xy_offset+fdy, z ],
        [ x1-xy_offset-rdx, y0+xy_offset+fdy, z ],
        [ x1-xy_offset-rdx, y1-xy_offset-bdy, z ],
        [ x0+xy_offset+ldx, y1-xy_offset-bdy, z ]
    ];
