// ease_shapes miscellaneous modules:
//      inside_corner
//      eased_cube_sideways
//      eased_cube_cutout_sideways
// filename: es_miscellaneous.scad

/* Copyright 2018-2020 Andrew H. Wakefield

    es_miscellaneous.scad is part of the eased_shapes.scad library

    This library is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

module inside_corner (

        // alternate ways of specifying the size
        // of the bevel or concave corner; size
        // overrides corner overrides r

        size = -1,
        corner = -1,
        r = -1,

        // alternate ways of specifying the length
        // of the corner piece; width overrides
        // length overrides h

        width = -1,
        length = -1,
        h = -1,

        // alternate ways of specifying the end
        // features; top/bottom override left/right
        // override front/back; top=right=back and
        // bottom=left=front

        bottom = -1,
        top = -1,
        left = -1,
        right = -1,
        front = -1,
        back = -1,

        // Type of feature [bottom/left,top/right,corner].
        // Note that corner can only be concave
        // or fillet; concave and fillet are treated
        // the same

        bevel = es_none,
        fillet = es_none,
        concave = es_none,

        // alternate mechanism for type of feature
        // [bevel,fillet,concave]. Same limitation
        // for corner_type - only concave or fillet.
        // Note that the highest priority feature
        // (concave) selected in ANY of bottom/left/
        // front or in top/right/back will be used

        bottom_type = es_none,
        top_type = es_none,
        left_type = es_none,
        right_type = es_none,
        front_type = es_none,
        back_type = es_none,
        corner_type = es_none,

        // alternate methods for specifiying axis
        // along which corner lies; on_z has the
        // highest priority, followed by on_y;
        // default is on_x

        on_x = true,
        on_y = false,
        on_z = false,

        direction = es_on_x,

        // rotation along the specified axis

        rot = 0,

        // if center = true, center along length; corner
        // is always at the center

        center = false
    )
{
    ch = es_get_priority_parameter (
              [ size, corner, r, 1 ] );
    cl = es_get_priority_parameter (
              [ width, length, h, 10 ] );

    ct = es_set_type(bevel[2],false,
        fillet[2]||concave[2],
        [ corner_type[0], false,
          corner_type[1] || corner_type[2] ],
        ch);

    bv = [ [ bevel[0], fillet[0], concave[0] ],
            bottom_type,
            left_type,
            front_type ];
    tv = [ [ bevel[1], fillet[1], concave[1] ],
            top_type,
            right_type,
            back_type ];

    init_b = es_get_priority_parameter (
        [ bottom, left, front, 0 ] );
    init_t = es_get_priority_parameter (
        [ top, right, back, 0 ] );

    b_type = es_set_type_v(init_b,bv);
    t_type = es_set_type_v(init_t,tv);
    c_type = ct[0] || ct[2] ? ct : [true,false,false];

    dir = es_set_dir(on_x,on_y,on_z,direction);

    bh = init_b > 0 && ! es_is_straight(init_b,b_type) ?
        init_b : 0;
    th = init_t > 0 && ! es_is_straight(init_t,t_type) ?
        init_t : 0;

    extra = max(th,bh,es_min_size);

    rx = ( dir[1] ? -90 : 0 );
    ry = ( dir[0] ? 90 : 0 );
    rz = rx+ry+rot;

    t = center ? -cl / 2 : 0;

    xt = dir[0] ? t : 0;
    yt = dir[1] ? t : 0;
    zt = dir[2] ? t : 0;

    translate([xt,yt,zt])
    rotate([rx,ry,0])
    rotate([0,0,rz])

    if ( c_type[0] ) {
        size = ( ch + extra * 2 ) * sqrt(2);

        difference() {
            translate([-extra,-extra,0])
            rotate([0,0,45])
            eased_cube([size,size,cl],
                top=th,top_type=t_type,
                bottom=bh,bottom_type=b_type,
                center_xy=true);

            translate([-size-extra,0,0])
            cube([size*2,size*2,cl*3],center=true);

            translate([0,-size-extra,0])
            cube([size*2,size*2,cl*3],center=true);
        }

    } else {
        size = ch + extra;
        
        difference() {
            translate([-extra,-extra,0])
            cube([size,size,cl]);
            
            translate([ch,ch,0])
            eased_cylinder_cutout(r=ch,h=cl,
                top=th,top_type=t_type,
                bottom=bh,bottom_type=b_type);
        }
    }
}

// helper modules to facilitate an eased_cube
// or eased_cube_cutout turned sideways so that
// the top and bottom become the left and right,
// and the corners become the top and bottom (tb)

module eased_cube_sideways
    (   size, tb, left, right,
        tb_type, left_type, right_type,
        center = false
    )
{
    xt = center ? 0 : size[0]/2;
    yt = center ? 0 : size[1]/2;
    zt = center ? 0 : size[2]/2;

    translate([xt,yt,zt])
    rotate([0,90,0])
    eased_cube([size[2],size[1],size[0]],
        top=left,top_type=left_type,
        bottom=right,bottom_type=right_type,
        corners=tb,corner_type=tb_type,
        center=true);
}

module eased_cube_cutout_sideways
    (   size, tb, left, right,
        tb_type, left_type, right_type,
        center = false
    )
{
    xt = center ? 0 : size[0]/2;
    yt = center ? 0 : size[1]/2;
    zt = center ? 0 : size[2]/2;

    translate([xt,yt,zt])
    rotate([0,90,0])
    eased_cube_cutout([size[2],size[1],size[0]],
        top=left,top_type=left_type,
        bottom=right,bottom_type=right_type,
        corners=tb,corner_type=tb_type,
        center=true);
}
